/*
 * Dump an ELF file's segments.  Make a cursory attempt at listing notes too.
 *
 * This software is being released under the GNU Public Licence.
 *
 * Copyright 1999, Jeremy Fitzhardinge <jeremy@goop.org>
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/param.h>

#include <elf.h>

#include "notes.h"

static const char *progname;

static const char *etype(int et) {
	switch(et) {
	case ET_EXEC:	return "ET_EXEC";
	case ET_CORE:	return "ET_CORE";
	case ET_DYN:	return "ET_DYN";
	case ET_REL:	return "ET_REL";
	case ET_NONE:	return "ET_NONE";
	default:	return "ET_???";
	}
}

static const char *ptype(int pt) {
	switch (pt) {
	case PT_NULL:	return "PT_NULL";
	case PT_LOAD:	return "PT_LOAD";
	case PT_DYNAMIC:return "PT_DYNAMIC";
	case PT_INTERP:	return "PT_INTERP";
	case PT_NOTE:	return "PT_NOTE";
	case PT_SHLIB:	return "PT_SHLIB";
	case PT_PHDR:	return "PT_PHDR";
	default:	return "PT_???";
	}
}

#define roundup(x, y)  ((((x)+((y)-1))/(y))*(y))

static void dumpnotes(char *data, int size)
{
	struct note *notes = noteparse(data, size, NULL);
	struct note *next;
	int i = 0;

	for(; notes; notes = next) {
		next = notes->next;

		printf("\t%d: \"%s\" type %d, %d bytes\n",
		       i++, notes->name, notes->type, notes->datasz);
		free(notes->name);
		free(notes->data);
		free(notes);
	}
}

int main(int argc, char **argv)
{
	char *inmap;
	int infd;
	int i;
	off_t len;
	Elf32_Ehdr *inhdr;
	Elf32_Phdr *inphdr;

	progname = argv[0];

	if (argc != 2) {
		fprintf(stderr, "Usage: %s file\n", progname);
		return 1;
	}

	if ((infd = open(argv[1], O_RDONLY)) == -1) {
		perror(argv[0]);
		return 1;
	}

	len = lseek(infd, 0, SEEK_END);
	
	if ((inmap = mmap(0, len, PROT_READ, MAP_PRIVATE, infd, 0)) == (char *)-1) {
		perror("mmap");
		return 1;
	}

	close(infd);

	inhdr = (Elf32_Ehdr *)inmap;
        inphdr= (Elf32_Phdr *)(inmap + inhdr->e_phoff);

	if (memcmp(&inhdr->e_ident[EI_MAG0], ELFMAG, SELFMAG) != 0) {
		fprintf(stderr, "%s: bad ELF magic\n", progname);
		return 1;
	}

	if (inhdr->e_ident[EI_CLASS] != ELFCLASS32) {
		fprintf(stderr, "%s: can only deal with 32-bit files\n",
			progname);
		return 1;
	}

	if (inhdr->e_ident[EI_DATA] != ELFDATA2LSB) {
		fprintf(stderr, "%s: BUG - can't deal with other endian\n",
			progname);
		return 1;
	}

	printf("ELF file: %d bytes\n", len);
	printf("\ttype = %s, cl=%d ds=%d\n", etype(inhdr->e_type), 
	       inhdr->e_ident[EI_CLASS], inhdr->e_ident[EI_DATA]);
	printf("\tmachine = %d\n", inhdr->e_machine);
	printf("\tversion = %d\n", inhdr->e_version);
	printf("\tentry = %08x\n", inhdr->e_entry);
	printf("\tphoff = %d\n", inhdr->e_phoff);
	printf("\tshoff = %d\n", inhdr->e_shoff);
	printf("\tflags = %x\n", inhdr->e_flags);
	printf("\tehsize = %d (sizeof Ehdr %d)\n", inhdr->e_ehsize, sizeof(Elf32_Ehdr));
	printf("\tphentsize = %d (sizeof Phdr %d)\n", inhdr->e_phentsize, sizeof(Elf32_Phdr));
	printf("\tphnum = %d\n", inhdr->e_phnum);
	printf("\tshentsize = %d (sizeof Shdr %d)\n", inhdr->e_shentsize, sizeof(Elf32_Shdr));
	printf("\tshnum = %d\n", inhdr->e_shnum);
	printf("\tshstrndx = %d\n", inhdr->e_shstrndx);

	for(i = 0; i < inhdr->e_phnum; i++) {
		Elf32_Phdr *ph = &inphdr[i];
		printf("%d: %-10s  va=%08x pa=%08x of=%08x fsz=%08x msz=%08x, al=%d, fl=%x\n",
		       i, ptype(ph->p_type), ph->p_vaddr, ph->p_paddr, ph->p_offset,
		       ph->p_filesz, ph->p_memsz, ph->p_align, ph->p_flags);
		switch(ph->p_type) {
		case PT_NOTE:
			dumpnotes(inmap + ph->p_offset, ph->p_filesz);
			break;
		}
	}

	return 0;
}
